#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <stropts.h>
#include <errno.h>
#include <sys/vuid_event.h>


#define LEFTBUTMASK 0x01
#define MIDDLEBUTMASK 0x04
#define RIGHTBUTMASK 0x02

void print8(unsigned char *buffer){
	printf("raw values: %2x:%2x:%2x:%2x\n",
	       buffer[0], buffer[1], buffer[2], buffer[3]
	       );

	
}

void print4(unsigned char *buffer){
	uint8_t button;
	int8_t x, y;
	int8_t wheel;
	char *wheelstat;
	int stats=0;

#ifdef PRINTRAW
	printf("raw values: %2x:%2x:%2x:%2x\n",
	       buffer[0], buffer[1], buffer[2], buffer[3]
	       );
#endif

	button=buffer[0];
	x=(int8_t)buffer[1];
	y=(int8_t)buffer[2];
	wheel=(int8_t)buffer[3];

	if(button&LEFTBUTMASK) printf("LEFT ");
	if(button&MIDDLEBUTMASK) printf("MIDDLE ");
	if(button&RIGHTBUTMASK) printf("RIGHT ");

	switch(wheel){
		case -1: wheelstat="down"; break;
		case 0: wheelstat=" --"; break;
		case 1: wheelstat="up"; break;
	}
	printf("x:%d y:%d wheel:%s\n",x,y,wheelstat);

}


int main(int argc, char *argv[]){
	int fd,bytes;
	char buffer[10];
	int readsize=4;

	if(argc<2){
		puts("Need a device name to hold open");
		exit(1);
	}
	fd=open(argv[1],O_RDONLY);
	if(fd<0){
		perror("could not open file");
		exit(1);
	}
	while(1){
	

#ifdef USE_NREAD
		if(ioctl(fd,I_NREAD,&bytes)!=0){
			perror("ioctl failed\n");
			return (errno);
		}
		if(bytes<=0){
			continue;
		}
		printf("bytes ready: %d\n",bytes);
#else
		/* hardcode for graphire testing */
		bytes=readsize;
#endif

		while(bytes>0){
			int readstat=0;
			readstat=read(fd,buffer,readsize);
			if(readstat!=readsize){
				int err=errno;
				perror("read of chars failed\n");
				printf("readstat==%d\n",readstat);
				exit(errno);
			}

			switch(readsize){
				case 8:
				print8(buffer);
				break;
				case 4:
				print4(buffer);
				break;
			}
			bytes-=readsize;

		}
	}

	/* NOTREACHED */
	puts("quitting");
	close(fd);
	
}
