/* This is a driver for a usbsm graphics tablet.
 * It was based on my PCIbase driver, from
 *  http://www.bolthole.com/solaris/drivers/
 *
 * It is being converted to a STREAMS type driver now.
 *    Philip Brown
 *
 *  @(#) usbsm.c 1.21@(#)
 */

#define USBDRV_MAJOR_VER        0
#define USBDRV_MINOR_VER        8

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/usb/usba.h>
  /* we dont actually use the hid routines.. we use the defines */
#include <sys/usb/clients/hid/hid.h>

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/file.h>

#include <sys/stream.h>
#include <stropts.h>
#include <sys/signal.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include "usbsm.h"

/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	name
#define DEVNODE_TYPE	DDI_NT_MOUSE /* see sunddi.h */
static char *name = "usbsm";

/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	usbsm_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);

static	int	usbsm_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);

static	int	usbsm_open(queue_t *q,dev_t *dev, int ofl, int otyp, cred_t *c);
static	int	usbsm_wput(queue_t *q,mblk_t *mblk);

static	int	usbsm_rput(queue_t *q,mblk_t *mp);
static	int	usbsm_rsrv(queue_t *q);

static	int	usbsm_close(queue_t *q, int flag, cred_t *cred_p);

static	int	usbsm_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static  int     usbsm_power(dev_info_t *, int, int);
static  int     usbsm_disconnect_callback(dev_info_t *dip);
static  int     usbsm_reconnect_callback(dev_info_t *dip);
static  void    usbsm_restore_device_state(dev_info_t *, usbsm_state_t *);
static  int     usbsm_pre_suspend_event_callback(dev_info_t *dip);
static  int     usbsm_init_power_mgmt(usbsm_state_t *);
static  void    usbsm_destroy_power_mgmt(usbsm_state_t *);
static  int     usbsm_post_resume_event_callback(dev_info_t *dip);

static  int     usbsm_get_descriptors(usbsm_state_t *statep);

static int debug_level;


/**********************************************************************/
/* STREAMS driver info stuff  */

static struct module_info stream_minfo = {
	0x0001,		/* module id number (?) */
	"usbsm",	/* module name */
	0,		/* min packet size */
	INFPSZ,		/* max packet size */
	512,		/* "high water mark" (?)*/
	128,		/* "low water mark" (?)*/
};

/* Queue for reading data */
static struct qinit stream_rinit = {
	usbsm_rput, /* qi_putp */
	usbsm_rsrv, /* qi_srvp */
	usbsm_open,
	usbsm_close,
	NULL, /* qi_qadmin */
	&stream_minfo,
	NULL /* qi_mstat */
};

/* Queue for writing data (to the dev) */
static struct qinit stream_winit = {
	/* Im not sure we should really have ANYTHING here.. */
	usbsm_wput, /* qi_put */ 
	NULL, /* qr_srv not needed */
	NULL, /* qi_qopen */
	NULL, /* qi_qclose */
	NULL, /* qi_qadmin */
	&stream_minfo,
	NULL /* qi_mstat */
};

static struct streamtab stream_rwinfo = {
	&stream_rinit,
	&stream_winit,
	NULL,		/* st_muxrinit */
	NULL		/* st_muxwinit */
};

/* end of STREAMS specific struct stuff  */
/**********************************************************************/



/* "char/block operations" */
/* Since we are a STREAMS dev, all the stuff that is NORMALLY referenced in 
 * here, is now referenced from the wierd STREAMs specific subtree.
 */
static struct cb_ops	usbsm_cb_ops = {
	nulldev,		/* open() */
	nulldev,		/* close() */
	nodev,			/* strategy()		*/
	nodev,			/* print routine	*/
	nodev,			/* no dump routine	*/
	nodev,			/* read() */
	nodev,			/* write() */
	nodev,			/* ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	&stream_rwinfo,		/* cb_stream struct */
	D_NEW | D_MP,		/* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops usbsm_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	usbsm_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	nulldev,		/* devo_probe */
	usbsm_attach,		/* devo_attach */
	usbsm_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&usbsm_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	usbsm_power,		/* devo_power */
};

/* Ignore this.
 * It is already declared in <sys/modctl.h>
 * It is only a filler for the next struct.
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"BOLTusbsm STREAM driver",
	&usbsm_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   usbsm_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * usbsm_glob_soft_handle;

static usb_event_t usbsm_events = {
        usbsm_disconnect_callback,
        usbsm_reconnect_callback,
        usbsm_pre_suspend_event_callback,
        usbsm_post_resume_event_callback,
};

/* Debugging support */
static uint_t   usbsm_errmask         = (uint_t)PRINT_MASK_ALL;
static uint_t   usbsm_errlevel        = USB_LOG_L4;
static uint_t   usbsm_instance_debug  = (uint_t)-1;
static uint_t   usbsm_show_label      = USB_ALLOW_LABEL;

static int      usblog_treelevel = 4; /* Valid values are -1 through 4. */

/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
usbsm_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	if(debug_level==0){
		return;
	}

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the usbsm_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See usbsm_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&usbsm_global_mutex, "usbsm: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct usbsm_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&usbsm_glob_soft_handle,
				sizeof(struct usbsm_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&usbsm_global_mutex);
		ddi_soft_state_fini(&usbsm_glob_soft_handle);
		cmn_err(CE_WARN, "usbsm: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	usbsm_debugprint("usbsm: _init returning SUCCESS\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&usbsm_glob_soft_handle);
		mutex_destroy(&usbsm_global_mutex);
	}
	return (ret);
}

/************************************************************/

/**********************************************************************/
/* This is probably interrupt context. 
 * mblk_t is a STREAMS style Message Block. Arrg.
 */
static void
usbsm_pipe_callback(usb_pipe_handle_t pipe, struct usb_intr_req *req)
{
	mblk_t *mblkptr = req->intr_data;
	struct usbsm_softstate *softstate=
			(struct usbsm_softstate*)req->intr_client_private;

	if(debug_level>1){
		usbsm_debugprint("usbsm_pipe_callback called\n");
	}

	if(mblkptr==NULL){
		printf("mblk==NULL\n");
		return;
	}

	if(softstate->rq==NULL){
		printf("ERROR in callback: softstate->rq==NULL?\n");
		return;
	}

	if(debug_level>=2){
	usbsm_debugprint("^usbsm_pipe_callback - calling usbsm_rput\n");
	}
	usbsm_rput(softstate->rq, mblkptr);
	req->intr_data = NULL;
	usb_free_intr_req(req);
}

/* This gets called for partial USB transfers.
 * EG: some error occurred.
 */
static void
usbsm_exception_callback(usb_pipe_handle_t pipe, struct usb_intr_req *req)
{
	mblk_t *mblkptr = req->intr_data;
	struct usbsm_softstate *softstate=
			(struct usbsm_softstate*)req->intr_client_private;

	usbsm_debugprint("usbsm_EXCEPTION_callback called: reason %d\n",
			 req->intr_exc_cb_flags);
	if(mblkptr)
	  usbsm_rput(softstate->rq, mblkptr);
	req->intr_data = NULL;
	usb_free_intr_req(req);
	usb_pipe_reset(softstate->dip, pipe, USB_FLAGS_SLEEP, NULL, NULL);
}

static int usbsm_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	struct usbsm_softstate *softstate;
	int returnval=DDI_FAILURE;
	dev_t devt=(dev_t)arg;
	int instance=DEVT_TO_INST(devt);
	
	usbsm_debugprint("usbsm: getinfo called\n");
	
	switch(infocmd){
	    case DDI_INFO_DEVT2DEVINFO:
		softstate=ddi_get_soft_state(usbsm_glob_soft_handle,instance);
		if(softstate==NULL){
			*result=NULL;
			break;
		}
		*result=softstate->dip;
		returnval=DDI_SUCCESS;
		break;
	    case DDI_INFO_DEVT2INSTANCE:
		*result=(void*)instance;
		returnval=DDI_SUCCESS;
		break;
	}

	return (returnval);
}




/* This routine can be called by usbsm_attach (if errors occur)
 * and usbsm_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void usbsm_cleanup(struct usbsm_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: usbsm_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	if( softstate->locks_initialized) {
	  usb_unregister_event_cbs(softstate->dip, &usbsm_events);
	  (void) usb_serialize_access(softstate->serial_cookiep, USB_WAIT, 0);
	  usb_release_access(softstate->serial_cookiep);
	  usbsm_destroy_power_mgmt(softstate);
	  usb_fini_serialization(softstate->serial_cookiep);
	}

	ddi_remove_minor_node(softstate->dip, name);

	usb_unregister_client(softstate->dip, softstate->reg);
	mutex_destroy(&softstate->instance_mutex);
	usb_free_log_handle(softstate->log_handle);
	ddi_prop_remove_all(softstate->dip);
	ddi_soft_state_free(usbsm_glob_soft_handle,instance);
}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here. We dont have registers to map, however.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in usbsm_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  usbsm_detach()
 */
static int usbsm_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	int status;
	struct usbsm_softstate *softstate;

	/* default debug level to **0**
	 * You can override in the .conf file.
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 0);

	usbsm_debugprint(" usbsm_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "usbsm: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "usbsm: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "usbsm: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	usbsm_debugprint("usbsm: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(usbsm_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "usbsm: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(usbsm_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "usbsm: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(usbsm_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}


	softstate->dip = dip;
	softstate->openflags = 0; /* redundant, but what the heck */
	softstate->log_handle = usb_alloc_log_handle(dip, name,
                &usbsm_errlevel, &usbsm_errmask,
                &usbsm_instance_debug, &usbsm_show_label, 0);
	
	usblog_treelevel = ddi_prop_get_int(
            DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS, "usblog_treelvl", -1);

	if ((status = usb_register_client(dip, USBDRV_VERSION,
            &softstate->reg, USB_REG_DESCR_IF, 0)) != USB_SUCCESS) {
                USB_DPRINTF_L0(PRINT_MASK_ALL, softstate->log_handle,
                    "attach: usb registration failed, error code:%d", status);
		usbsm_cleanup(softstate);
		return DDI_FAILURE;
        }

	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER,
		   softstate->reg->dev_iblock_cookie);

	softstate->serial_cookiep = usb_init_serialization(dip, 0);
	softstate->locks_initialized = B_TRUE;

        if (usblog_treelevel >= 0) {
                (void) usb_log_descr_tree(softstate->reg,
                    softstate->log_handle, usblog_treelevel, PRINT_MASK_ATT);
        }

	if (usbsm_get_descriptors(softstate) != USB_SUCCESS) {
                USB_DPRINTF_L0(PRINT_MASK_ATT, softstate->log_handle,
                    "usb get descriptors failed");
		usbsm_cleanup(softstate);
		return DDI_FAILURE;
        }

        /* we no longer need the descriptor tree */
        usb_free_descr_tree(dip, softstate->reg);

	usbsm_debugprint(" maxtransfer sizeval=%u",
	                  softstate->endpoint.wMaxPacketSize);


	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */

	/* XXX I should really check the dev type, and change the
	 * name based on that stuff
	 */
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* There are no PCI-style interrupts, since we dont get to directly
	 * connect to a device.
	 * but an interrupt handler of sorts gets set up in open()
	 */

	softstate->dev_state = USB_DEV_ONLINE;
	if(usbsm_init_power_mgmt(softstate) != USB_SUCCESS) {
	  USB_DPRINTF_L3(PRINT_MASK_ATT, softstate->log_handle,
			 "attach: Error initializing power mgmt");
	}
	if (usb_register_event_cbs(dip, &usbsm_events, 0) != USB_SUCCESS) {
		usbsm_cleanup(softstate);
		return DDI_FAILURE;
	}

	ddi_report_dev(dip);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}

/* we dont actualy USE this for anything.
 * At least, not right now.
 */
static	int	usbsm_wput(queue_t *q,mblk_t *mblk){
	printf("WPUT called?\n");
	return 0;
}


/* usbsm_rput:   precursor to the usbsm_rsrv routine.
 * We put stuff on OUR 'STREAM' queue here.
 * Then the rsrv routine moves the stuff over to the UPSTREAM queue,
 * aka the 'stream head', which the user talks to.
 */
static int
usbsm_rput(queue_t *q,mblk_t *mp)
{
	if(debug_level >=2){
		usbsm_debugprint("^rput called. Placing %p on our queue\n",mp);
	}
	putq(q,mp);
	return 0;
}


/* STREAMS 'service' proceedure, for the read queue.
 * called by the kernel STREAMS scheduler.
 * Basically, just takes all the mouse events we have queued up internally,
 * and shoves them up the stream towards user-level.
 */
static int
usbsm_rsrv(queue_t *q)
{
	mblk_t *mq;

	if(debug_level>=2){
		usbsm_debugprint("^rsrv called\n");
	}

	while((mq=getq(q)) != NULL){
		if(!canputnext(q)){
			putbq(q,mq);
			break;
		}
		putnext(q,mq);
	}
	return 0;
}


static int
usbsm_open(queue_t *rq,dev_t *dev, int openflags, int sflag, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	usb_pipe_policy_t policy;
	struct usbsm_softstate *softstate;
	usb_pipe_handle_t pipe_handle;
	usb_intr_req_t *req;
	int rval = USB_SUCCESS;

	ASSERT(rq); /* bit of sun paranoia */

	usbsm_debugprint("usbsm: open for instance %d\n", instance);

	/* We dont handle any special stream options
	 * like MODOPEN or CLONEOPEN
	 */
	if (sflag){
		usbsm_debugprint("Cannot open with STREAM flags set\n");
		return EINVAL;
	}
	if (openflags & FWRITE){
		usbsm_debugprint("Cannot open device for write\n");
		return EINVAL;
	}


	softstate=ddi_get_soft_state(usbsm_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is NOT technically a failure: it is sometimes expected.
		 * It is called "deferred attach"
		 */
		return (ENXIO);
	}

	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!=0){
		usbsm_debugprint("usbsm: device already opened\n");
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
	}
	softstate->openflags=USBMS_OPEN_HELD;
	mutex_exit(&softstate->instance_mutex);

	/*
         * This is in place so that a disconnect or CPR doesn't interfere with
         * pipe opens.
         */
	if (usb_serialize_access(softstate->serial_cookiep,
				 USB_WAIT_SIG, 0) == 0) {
		mutex_enter(&softstate->instance_mutex);
                softstate->openflags=0;
		mutex_exit(&softstate->instance_mutex);
                return (EINTR);
        }

	/* raise power */
        (void) pm_busy_component(softstate->dip, 0);
        (void) pm_raise_power(softstate->dip, 0, USB_DEV_OS_FULL_PWR);

	bzero(&policy,sizeof(policy));
	policy.pp_max_async_reqs = 2;

	if((rval=usb_pipe_open(softstate->dip,&softstate->endpoint, &policy,
			 USB_FLAGS_SLEEP, &pipe_handle)) !=USB_SUCCESS){
		USB_DPRINTF_L0(PRINT_MASK_ATT, softstate->log_handle,
				 "Error opening intr pipe: status = %d", rval);
		mutex_enter(&softstate->instance_mutex);
		softstate->openflags=0;
		mutex_exit(&softstate->instance_mutex);
		(void) pm_idle_component(softstate->dip, 0);
		usb_release_access(softstate->serial_cookiep);
		return (ENXIO);
	}

	rq->q_ptr = WR(rq)->q_ptr = softstate;
	mutex_enter(&softstate->instance_mutex);
	softstate->usb_pipe_handle = pipe_handle;
	softstate->rq = rq;
	mutex_exit(&softstate->instance_mutex);
	qprocson(rq);

	req = usb_alloc_intr_req(softstate->dip, 0, USB_FLAGS_SLEEP);
	/* this is a default anyway, should we still set it? */
	req->intr_len = softstate->endpoint.wMaxPacketSize;
	req->intr_client_private = (usb_opaque_t)softstate;
	req->intr_attributes = USB_ATTRS_SHORT_XFER_OK|USB_ATTRS_AUTOCLEARING;
	req->intr_cb = usbsm_pipe_callback;
	req->intr_exc_cb = usbsm_exception_callback;
	
	if((rval=usb_pipe_intr_xfer(softstate->usb_pipe_handle, req,
				  USB_FLAGS_SLEEP)) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling FAILED: %d", rval);
		usb_pipe_close(softstate->dip, softstate->usb_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL );

		req->intr_client_private = NULL; /* just in case */
		usb_free_intr_req(req);
		mutex_enter(&softstate->instance_mutex);
		softstate->usb_pipe_handle=NULL;
		softstate->openflags=0;
		mutex_exit(&softstate->instance_mutex);
		(void) pm_idle_component(softstate->dip, 0);
		usb_release_access(softstate->serial_cookiep);
		return ENXIO;
	}

	usbsm_debugprint("usbsm_open called start_polling successfully\n");

	(void) pm_idle_component(softstate->dip, 0);
	usb_release_access(softstate->serial_cookiep);
   	return 0;
}


static int usbsm_close(queue_t *q, int flag, cred_t *cred_p){
	struct usbsm_softstate *softstate;

   	usbsm_debugprint("usbsm: close called\n");
	/* Okay, extreme paranoia for my first STREAMS dev */
	if(q==NULL) return ENXIO;
	if(q->q_ptr==NULL) return ENXIO;

	/* XXX in theory, might want to do additional stuff like
	 *  quntimeout(), qunbufcall(), .... ?
	 */
	softstate=(struct usbsm_softstate*)q->q_ptr;
	(void) usb_serialize_access(softstate->serial_cookiep, USB_WAIT, 0);
	(void) pm_busy_component(softstate->dip, 0);
        (void) pm_raise_power(softstate->dip, 0, USB_DEV_OS_FULL_PWR);

	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!= USBMS_OPEN_HELD){
		cmn_err(CE_WARN,"ERROR: usbsm_close: pipe not open");
		mutex_exit(&softstate->instance_mutex);
		usb_release_access(softstate->serial_cookiep);
		return ENXIO;
		
	}
	softstate->openflags= USBMS_OPEN_ENDING;
	mutex_exit(&softstate->instance_mutex);

	qprocsoff(q);

	usb_pipe_stop_intr_polling(softstate->usb_pipe_handle, USB_FLAGS_SLEEP);
	usb_pipe_close(softstate->dip, softstate->usb_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL );

	mutex_enter(&softstate->instance_mutex);
	softstate->usb_pipe_handle=NULL;
	softstate->openflags=0;
	mutex_exit(&softstate->instance_mutex);

	q->q_ptr=WR(q)->q_ptr=NULL;

	(void) pm_idle_component(softstate->dip, 0);
	usb_release_access(softstate->serial_cookiep);
   	return (0);
}

static int usbsm_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct usbsm_softstate *softstate;
   	instance=ddi_get_instance(dip);

	usbsm_debugprint("usbsm: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(usbsm_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"usbsm_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	usbsm_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}

static int
usbsm_get_descriptors(usbsm_state_t *statep)
{
        int                     interface;
        usb_client_dev_data_t   *dev_data =
                                    statep->reg;
        usb_alt_if_data_t       *altif_data;
        usb_cfg_data_t          *cfg_data;
        usb_ep_data_t           *ep_data;
        dev_info_t              *dip = statep->dip;
        int                     alt, rval;

        ASSERT(!mutex_owned(&statep->instance_mutex));

        interface = dev_data->dev_curr_if;
        cfg_data = dev_data->dev_curr_cfg;
        alt = 0;

        /*
         * find the endpoint descriptors. There should be a intr-in endpoint.
         */
        if ((ep_data = usb_get_ep_data(dip, dev_data, interface, alt,
            USB_EP_ATTR_INTR, USB_EP_DIR_IN)) != NULL) {
                mutex_enter(&statep->instance_mutex);
                statep->endpoint = ep_data->ep_descr;
                mutex_exit(&statep->instance_mutex);
        }

        return (USB_SUCCESS);
}

/*
 * disconnect_callback:
 *      Called when device hotplug-removed.
 *      Close pipes. (This does not attempt to contact device.)
 *      Set state to DISCONNECTED
 */
static int
usbsm_disconnect_callback(dev_info_t *dip)
{
   int instance = ddi_get_instance(dip);
   usbsm_state_t *statep = 
		ddi_get_soft_state(usbsm_glob_soft_handle, instance);
   struct iocblk *mctlmsg;
   mblk_t *mp;
   queue_t *q;

   USB_DPRINTF_L4(PRINT_MASK_ATT, statep->log_handle, "disconnect: enter");

   (void) usb_serialize_access(statep->serial_cookiep, USB_WAIT, 0);

   /*
    * Save any state of device or IO in progress required by
    * usbsm_restore_device_state for proper device "thawing" later.
    */
   mutex_enter(&statep->instance_mutex);
   switch(statep->dev_state) {
   case USB_DEV_ONLINE:
   case USB_DEV_PWRED_DOWN:
     statep->dev_state = USB_DEV_DISCONNECTED;
     if(statep->openflags) {
       q = statep->rq;
       mutex_exit(&statep->instance_mutex);
       if (canputnext(q)) {
	  mp = allocb((int)sizeof (struct iocblk), NULL);
	  mp->b_datap->db_type = M_CTL;
	  mctlmsg = (struct iocblk *)mp->b_datap->db_base;
	  mctlmsg->ioc_cmd = HID_DISCONNECT_EVENT;
	  mctlmsg->ioc_count = 0;
	  (void) putnext(q, mp);
       }
       usb_pipe_stop_intr_polling(statep->usb_pipe_handle, USB_FLAGS_SLEEP);
     } else {
       mutex_exit(&statep->instance_mutex);
     }
     break;
   case USB_DEV_SUSPENDED:
     /* we remain suspended */
     mutex_exit(&statep->instance_mutex);
     break;
   default:
     USB_DPRINTF_L2(PRINT_MASK_USE, statep->log_handle, "usbsm_disconnect_callback: Illegal dev state: %d", statep->dev_state);
     mutex_exit(&statep->instance_mutex);
     break;
   }

   usb_release_access(statep->serial_cookiep);

   return (USB_SUCCESS);
}

static int usbsm_reconnect_callback(dev_info_t *dip)
{
   int instance = ddi_get_instance(dip);
   usbsm_state_t *statep =
                     ddi_get_soft_state(usbsm_glob_soft_handle, instance);

   USB_DPRINTF_L4(PRINT_MASK_ATT, statep->log_handle, "reconnect: enter");

   (void) usb_serialize_access(statep->serial_cookiep, USB_WAIT, 0);
   mutex_enter(&statep->instance_mutex);
   if(statep->dev_state == USB_DEV_ONLINE) {
	mutex_exit(&statep->instance_mutex);
	return USB_SUCCESS;
   }
   usbsm_restore_device_state(dip, statep);
   mutex_exit(&statep->instance_mutex);
   (void) usb_release_access(statep->serial_cookiep);

   return USB_SUCCESS;
}

static void usbsm_restore_device_state(dev_info_t *dip,
					  usbsm_state_t *statep)
{
  int rval;
  queue_t         *q;
  struct iocblk   *mctlmsg;
  mblk_t          *mp;

  USB_DPRINTF_L4(PRINT_MASK_ATT, statep->log_handle,
            "usbsm_restore_device_state: enter");

  ASSERT((statep->dev_state == USB_DEV_DISCONNECTED) ||
            (statep->dev_state == USB_DEV_SUSPENDED));

  if ((statep->pm != NULL) &&
            (statep->pm->usbsm_wakeup_enabled)) {
                mutex_exit(&statep->instance_mutex);
                (void) pm_busy_component(statep->dip, 0);
                (void) pm_raise_power(statep->dip, 0, USB_DEV_OS_FULL_PWR);
  } else {
    mutex_exit(&statep->instance_mutex);
  }

  /* Check if we are talking to the same device */
  if (usb_check_same_device(dip, statep->log_handle,
    USB_LOG_L0, PRINT_MASK_ATT, USB_CHK_ALL, NULL) != USB_SUCCESS) {
      mutex_enter(&statep->instance_mutex);
      /* change the device state from suspended to disconnected */
      statep->dev_state = USB_DEV_DISCONNECTED;
      (void) pm_idle_component(statep->dip, 0);
      return;
  }

  if ((statep->pm != NULL) && (statep->pm->usbsm_wakeup_enabled)) {

    /* Failure here means device disappeared again. */
    if (usb_handle_remote_wakeup(dip, USB_REMOTE_WAKEUP_ENABLE) != USB_SUCCESS)
    {
      USB_DPRINTF_L0(PRINT_MASK_ALL, statep->log_handle,
                            "Warning: device may or may not be accessible. "
                            "Please verify reconnection");
    }
  }

  if(statep->openflags) { /* device was open before disconnect */
    usb_intr_req_t *req;
    int rval;

    req = usb_alloc_intr_req(statep->dip, 0, USB_FLAGS_SLEEP);
    req->intr_len = statep->endpoint.wMaxPacketSize;
    req->intr_client_private = (usb_opaque_t)statep;
    req->intr_attributes = USB_ATTRS_SHORT_XFER_OK|USB_ATTRS_AUTOCLEARING;
    req->intr_cb = usbsm_pipe_callback;
    req->intr_exc_cb = usbsm_exception_callback;
    if((rval=usb_pipe_intr_xfer(statep->usb_pipe_handle, req,
                                  USB_FLAGS_SLEEP)) != USB_SUCCESS)
    {
	USB_DPRINTF_L3(PRINT_MASK_ATT, statep->log_handle,
                            "usbsm_restore_device_state:"
                            "unable to restart intr pipe poll"
                            " rval = %d ", rval);
	(void) pm_idle_component(dip, 0);
        mutex_enter(&statep->instance_mutex);
	statep->dev_state = USB_DEV_DISCONNECTED;
        return;
    }
    q = statep->rq;
    if (canputnext(q)) {
      mp = allocb((int)sizeof (struct iocblk), NULL);
      mp->b_datap->db_type = M_CTL;
      mctlmsg = (struct iocblk *)mp->b_datap->db_base;
      mctlmsg->ioc_cmd = HID_CONNECT_EVENT;
      mctlmsg->ioc_count = 0;
      (void) putnext(q, mp);
    }
  }
  mutex_enter(&statep->instance_mutex);
  statep->dev_state = USB_DEV_ONLINE;
  mutex_exit(&statep->instance_mutex);

  (void) pm_idle_component(statep->dip, 0);

  USB_DPRINTF_L4(PRINT_MASK_ATT, statep->log_handle,
            "usbsm_restore_device_state: End");

  mutex_enter(&statep->instance_mutex);

  return;
}

static int
usbsm_pre_suspend_event_callback(dev_info_t *dip)
{
        int             instance = ddi_get_instance(dip);
        usbsm_state_t *statep = ddi_get_soft_state(usbsm_glob_soft_handle,
                                                        instance);
        int             old_dev_state;

        USB_DPRINTF_L4(PRINT_MASK_ATT, statep->log_handle,
            "pre_suspend_event_callback enter");

        /* Serialize to prevent races with detach, open, device access. */
        (void) usb_serialize_access(statep->serial_cookiep,
                                USB_WAIT, 0);

        (void) pm_busy_component(statep->dip, 0);
        (void) pm_raise_power(statep->dip, 0,
                                USB_DEV_OS_FULL_PWR);

        /* Access device here to clean it up. */

        /*
         * Set dev_state to suspended so other driver threads don't start any
         * new I/O.  In a real driver, there may be draining of requests done
         * afterwards, and we don't want the draining to compete with new
         * requests being queued.
         */
        mutex_enter(&statep->instance_mutex);
        old_dev_state = statep->dev_state;
        statep->dev_state = USB_DEV_SUSPENDED;

        /* Don't suspend if the device is open. */
        if (statep->openflags != 0) {
                USB_DPRINTF_L0(PRINT_MASK_ATT, statep->log_handle,
                    "pre_suspend_event_callback: "
                    "Device is open.  Can't suspend.");
                statep->dev_state = old_dev_state;
                mutex_exit(&statep->instance_mutex);
                usb_release_access(statep->serial_cookiep);
                (void) pm_idle_component(statep->dip, 0);
                usb_fail_checkpoint(dip, 0);

                return (USB_FAILURE);
        }
        /*
         * Save any state of device required by usbsm_restore_device_state
         * for proper device "thawing" later.
         */

        mutex_exit(&statep->instance_mutex);
        usb_release_access(statep->serial_cookiep);
        (void) pm_idle_component(statep->dip, 0);

        USB_DPRINTF_L4(PRINT_MASK_ATT, statep->log_handle,
            "pre_suspend_event_callback: success");

        return (USB_SUCCESS);
}

static int usbsm_post_resume_event_callback(dev_info_t *dip)
{
        int             instance = ddi_get_instance(dip);
        usbsm_state_t *statep = ddi_get_soft_state(usbsm_glob_soft_handle,
                                                        instance);

        USB_DPRINTF_L4(PRINT_MASK_ALL, statep->log_handle,
            "post_resume_event_callback: enter");

        /*
         * Needed as power-up state of device is "unknown" to system.
         * Need to sync.
         */
        (void) pm_raise_power(statep->dip, 0, USB_DEV_OS_FULL_PWR);

        /*
         * Do nothing if already ONLINE.  There are no requests to thaw, etc.
         * We're ready to go.
         */
        mutex_enter(&statep->instance_mutex);
        if (statep->dev_state != USB_DEV_ONLINE) {
                (void) usbsm_restore_device_state(dip, statep);
        }
        mutex_exit(&statep->instance_mutex);

        return (USB_SUCCESS);
}

/*
 * usbsm_power :
 *	Power entry point, the workhorse behind pm_raise_power, pm_lower_power,
 *	usb_req_raise_power and usb_req_lower_power.
 */
/* ARGSUSED */
static int
usbsm_power(dev_info_t *dip, int comp, int level)
{
	usbsm_state_t	*statep;
	usbsm_power_t	*pm;
	int	rval = USB_FAILURE;

	statep = (usbsm_state_t *)ddi_get_soft_state(usbsm_glob_soft_handle,
						ddi_get_instance(dip));

	USB_DPRINTF_L4(PRINT_MASK_PM, statep->log_handle,
	    "usbsm_power: enter: level = %d", level);

	(void) usb_serialize_access(statep->serial_cookiep,
					USB_WAIT, 0);

	mutex_enter(&statep->instance_mutex);

	/*
	 * If we are disconnected/suspended, return success. Note that if we
	 * return failure, bringing down the system will hang when
	 * PM tries to power up all devices
	 */
	if ((statep->dev_state == USB_DEV_DISCONNECTED) ||
		(statep->dev_state == USB_DEV_SUSPENDED)) {

		USB_DPRINTF_L4(PRINT_MASK_PM, statep->log_handle,
		    "usbsm_power: disconnected/suspended ",
		    "dev_state=%d", statep->dev_state);
		rval = USB_SUCCESS;

		goto done;
	}

	pm = statep->pm;

	/* Check if we are transitioning to a legal power level */
	if (USB_DEV_PWRSTATE_OK(pm->usbsm_pwr_states, level)) {
		USB_DPRINTF_L4(PRINT_MASK_PM, statep->log_handle,
		    "usbsm_power: illegal power level = %d "
		    "pwr_states: %x", level, pm->usbsm_pwr_states);

		goto done;
	}

	switch (level) {
	case USB_DEV_OS_PWR_OFF :
		if (statep->dev_state == USB_DEV_ONLINE) {
			statep->dev_state = USB_DEV_PWRED_DOWN;
			statep->pm->usbsm_current_power =
			    USB_DEV_OS_PWR_OFF;
			mutex_exit(&statep->instance_mutex);

			/* Issue USB D3 command to the device here */
			rval = usb_set_device_pwrlvl3(dip);
			mutex_enter(&statep->instance_mutex);
		} else {
			rval = USB_SUCCESS;
		}
		break;

	case USB_DEV_OS_FULL_PWR :
		/*
		 * PM framework tries to put us in full power during system
		 * shutdown.
		 */
		statep->dev_state = USB_DEV_ONLINE;
		statep->pm->usbsm_current_power =
		    USB_DEV_OS_FULL_PWR;
		mutex_exit(&statep->instance_mutex);
		rval = usb_set_device_pwrlvl0(dip);
		mutex_enter(&statep->instance_mutex);
		break;

	/* Levels 1 and 2 are not supported by this driver to keep it simple. */
	default:
		USB_DPRINTF_L0(PRINT_MASK_PM, statep->log_handle,
		    "usbsm_power: power level %d not supported", level);
		break;
	}
done:
	mutex_exit(&statep->instance_mutex);
	usb_release_access(statep->serial_cookiep);

	return ((rval == USB_SUCCESS) ? DDI_SUCCESS : DDI_FAILURE);
}


/*
 * usbsm_init_power_mgmt:
 *	Initialize power management and remote wakeup functionality.
 *	No mutex is necessary in this function as it's called only by attach.
 */
static int
usbsm_init_power_mgmt(usbsm_state_t *statep)
{
	int 		rval = USB_FAILURE;
	usbsm_power_t *statepm;
	dev_info_t 	*dip = statep->dip;

	USB_DPRINTF_L4(PRINT_MASK_PM, statep->log_handle,
	    "init_power_mgmt enter");

	if (usb_is_pm_enabled(dip) == USB_SUCCESS) {
		uint_t		pwr_states;

		/* Allocate the state structure */
		statepm = kmem_zalloc(sizeof (usbsm_power_t), KM_SLEEP);
		statep->pm = statepm;
		statepm->usbsm_state = statep;
		statepm->usbsm_wakeup_enabled = B_FALSE;
		statepm->usbsm_pm_capabilities = 0;
		statepm->usbsm_current_power = USB_DEV_OS_FULL_PWR;

		/*
		 * If remote wakeup is not available you may not want to do
		 * power management.
		 */
		if (usb_handle_remote_wakeup(dip,
		    USB_REMOTE_WAKEUP_ENABLE) == USB_SUCCESS) {

			if ((rval =
			    usb_create_pm_components(dip, &pwr_states)) ==
			    USB_SUCCESS) {

				USB_DPRINTF_L4(PRINT_MASK_PM,
				    statep->log_handle,
				    "usbsm_init_power_mgmt: "
				    "created PM components");

				statepm->usbsm_pwr_states =
							(uint8_t)pwr_states;
				statepm->usbsm_wakeup_enabled = B_TRUE;
				if (pm_raise_power(dip, 0,
				    USB_DEV_OS_FULL_PWR) != DDI_SUCCESS) {
					USB_DPRINTF_L1(PRINT_MASK_PM,
					    statep->log_handle,
					    "usbsm_init_power_mgmt: "
					    "raising power failed");
					rval = USB_FAILURE;
				}
			} else {
				USB_DPRINTF_L2(PRINT_MASK_PM,
				    statep->log_handle,
				    "usbsm_init_power_mgmt: "
				    "create_pm_compts failed");
			}
		} else {
			USB_DPRINTF_L2(PRINT_MASK_PM,
			    statep->log_handle,
			    "usbsm_init_power_mgmt: "
			    "failure enabling remote wakeup");
		}
	} else {
		USB_DPRINTF_L3(PRINT_MASK_PM, statep->log_handle,
		    "usbsm_init_power_mgmt: pm not enabled");
	}

	USB_DPRINTF_L4(PRINT_MASK_PM, statep->log_handle,
	    "usbsm_init_power_mgmt: end");

	return (rval);
}


/*
 * usbsm_destroy_power_mgmt:
 *	Shut down and destroy power management and remote wakeup functionality.
 */
static void
usbsm_destroy_power_mgmt(usbsm_state_t *statep)
{
	USB_DPRINTF_L4(PRINT_MASK_ALL, statep->log_handle,
	    "destroy_power_mgmt enter");

	if (statep->pm) {
		(void) pm_busy_component(statep->dip, 0);
		if (statep->pm->usbsm_wakeup_enabled) {

			mutex_enter(&statep->instance_mutex);
			if (statep->dev_state !=
			    USB_DEV_DISCONNECTED) {
				int rval;

				mutex_exit(&statep->instance_mutex);

				(void) pm_raise_power(statep->dip, 0,
				    USB_DEV_OS_FULL_PWR);

				if ((rval = usb_handle_remote_wakeup(
				    statep->dip,
				    USB_REMOTE_WAKEUP_DISABLE)) !=
				    USB_SUCCESS) {
					USB_DPRINTF_L4(PRINT_MASK_PM,
					    statep->log_handle,
					    "usbsm_destroy_power_mgmt: "
					    "disabling rmt wakeup: rval = %d",
					    rval);
				}
			} else {
				mutex_exit(&statep->instance_mutex);
			}

			/*
			 * Since remote wakeup is disabled now,
			 * no one can raise power
			 * and get to device once power is lowered here.
			 */
		}

		(void) pm_lower_power(statep->dip, 0,
		    USB_DEV_OS_PWR_OFF);
		(void) pm_idle_component(statep->dip, 0);
		kmem_free(statep->pm, sizeof (usbsm_power_t));
		statep->pm = NULL;
	}
}


