#ifndef _USBSM_H
#define _USBSM_H

/* power support */
typedef struct usbsm_power  {
  void *usbsm_state;
  boolean_t usbsm_wakeup_enabled;
  uint8_t usbsm_pwr_states;
  uint8_t usbsm_pm_capabilities;
  boolean_t usbsm_raise_power;
  uint8_t usbsm_current_power;
} usbsm_power_t;

/* This gets filled out on a per-instance basis by usbsm_attach */
typedef struct usbsm_softstate {
        dev_info_t *dip; /* you always need this cached */
        kmutex_t instance_mutex;

        uint8_t lastdata[8];

        usb_client_dev_data_t *reg;
        usb_ep_descr_t endpoint;
        int openflags; /* tell if we have been open()ed or not */
        usb_pipe_handle_t usb_pipe_handle;

        queue_t *rq, *wq;

        /* since we are not a streams driver we have to make our
         * own buffer/queue
         * "end" is where to add new data. "start" is where read()
         * should take data from
         */
        unsigned char *readbuffer;
        unsigned char *readbuf_start,*readbuf_end;

        boolean_t locks_initialized;
        int dev_state;                  /* USB states */
        usb_log_handle_t log_handle;    /* USB logging */
        usb_serialization_t serial_cookiep;     /* USB serialization */
        usbsm_power_t *pm;              /* power management */
} usbsm_state_t;

/* USB logging aid */
#define PRINT_MASK_ALL          0xFFFFFFFF
#define PRINT_MASK_PM           0x00000001
#define PRINT_MASK_ATT          0x00000002
#define PRINT_MASK_USE          0x00000004

/* defines used on the "openflags" var in softstate*/
#define USBMS_OPEN_HELD         0x01
#define USBMS_OPEN_ENDING       0x02
#define USBMS_OPEN_SHUTDOWN     0x03 /* flag "shutting down" case */

#endif
